import pytest
import numpy as np
from numpy.testing import assert_,  run_module_suite

from qutip import (ssesolve, destroy, coherent, mesolve, fock, qeye,
                   parallel_map, photocurrent_sesolve, num)

def f(t, args):
    return args["a"] * t


@pytest.mark.slow
def test_ssesolve_homodyne_methods():
    "Stochastic: ssesolve: homodyne methods with single jump operator"

    def arccoth(x):
        return 0.5*np.log((1.+x)/(x-1.))

    th = 0.1 # Interaction parameter
    alpha = np.cos(th)
    beta = np.sin(th)
    gamma = 1.

    N = 30                 # number of Fock states
    Id = qeye(N)
    a = destroy(N)
    s = 0.5*((alpha+beta)*a + (alpha-beta)*a.dag())
    x = (a + a.dag()) * 2**-0.5
    H = Id + gamma * a * a.dag()
    sc_op = [s]
    e_op = [x, x*x]
    rho0 = fock(N,0)      # initial vacuum state

    T = 6.                   # final time
    # number of time steps for which we save the expectation values
    N_store = 200
    Nsub = 10
    tlist = np.linspace(0, T, N_store)
    ddt = (tlist[1]-tlist[0])

    #### No analytic solution for ssesolve, taylor15 with 500 substep
    sol = ssesolve(H, rho0, tlist, sc_op, e_op,
                   nsubsteps=1000, method='homodyne', solver='taylor1.5')
    y_an = (sol.expect[1]-sol.expect[0]*sol.expect[0].conj())


    list_methods_tol = [['euler-maruyama', 3e-2],
                        ['pc-euler', 5e-3],
                        ['pc-euler-2', 5e-3],
                        ['platen', 5e-3],
                        ['milstein', 5e-3],
                        ['milstein-imp', 5e-3],
                        ['taylor1.5', 5e-4],
                        ['taylor1.5-imp', 5e-4],
                        ['explicit1.5', 5e-4],
                        ['taylor2.0', 5e-4]]
    for n_method in list_methods_tol:
        sol = ssesolve(H, rho0, tlist, sc_op, e_op,
                       nsubsteps=Nsub, method='homodyne', solver=n_method[0])
        sol2 = ssesolve(H, rho0, tlist, sc_op, e_op, store_measurement=0,
                       nsubsteps=Nsub, method='homodyne', solver=n_method[0],
                       noise = sol.noise)
        sol3 = ssesolve(H, rho0, tlist, sc_op, e_op,
                        nsubsteps=Nsub*10, method='homodyne',
                        solver=n_method[0], tol=1e-8)
        err = 1/T * np.sum(np.abs(y_an - \
                    (sol.expect[1]-sol.expect[0]*sol.expect[0].conj())))*ddt
        err3 = 1/T * np.sum(np.abs(y_an - \
                    (sol3.expect[1]-sol3.expect[0]*sol3.expect[0].conj())))*ddt
        print(n_method[0], ': deviation =', err, err3,', tol =', n_method[1])
        assert_(err < n_method[1])
        # 5* more substep should decrease the error
        assert_(err3 < err)
        # just to check that noise is not affected by ssesolve
        assert_(np.all(sol.noise == sol2.noise))
        assert_(np.all(sol.expect[0] == sol2.expect[0]))

    sol = ssesolve(H, rho0, tlist[:2], sc_op, e_op, noise=10, ntraj=2,
                    nsubsteps=Nsub, method='homodyne', solver='euler',
                    store_measurement=1)
    sol2 = ssesolve(H, rho0, tlist[:2], sc_op, e_op, noise=10, ntraj=2,
                    nsubsteps=Nsub, method='homodyne', solver='euler',
                    store_measurement=0)
    sol3 = ssesolve(H, rho0, tlist[:2], sc_op, e_op, noise=11, ntraj=2,
                    nsubsteps=Nsub, method='homodyne', solver='euler')
    # sol and sol2 have the same seed, sol3 differ.
    assert_(np.all(sol.noise == sol2.noise))
    assert_(np.all(sol.noise != sol3.noise))
    assert_(not np.all(sol.measurement[0] == 0.+0j))
    assert_(np.all(sol2.measurement[0] == 0.+0j))
    sol = ssesolve(H, rho0, tlist[:2], sc_op, e_op, noise=np.array([1,2]),
                   ntraj=2, nsubsteps=Nsub, method='homodyne', solver='euler')
    sol2 = ssesolve(H, rho0, tlist[:2], sc_op, e_op, noise=np.array([2,1]),
                   ntraj=2, nsubsteps=Nsub, method='homodyne', solver='euler')
    # sol and sol2 have the seed of traj 1 and 2 reversed.
    assert_(np.all(sol.noise[0,:,:,:] == sol2.noise[1,:,:,:]))
    assert_(np.all(sol.noise[1,:,:,:] == sol2.noise[0,:,:,:]))


def test_ssesolve_photocurrent():
    "Stochastic: photocurrent_sesolve"
    tol = 0.01

    N = 4
    gamma = 0.25
    ntraj = 25
    nsubsteps = 100
    a = destroy(N)

    H = [[a.dag() * a,f]]
    psi0 = coherent(N, 0.5)
    sc_ops = [np.sqrt(gamma) * a, np.sqrt(gamma) * a*0.5]
    e_ops = [a.dag() * a, a + a.dag(), (-1j)*(a - a.dag())]

    times = np.linspace(0, 2.5, 50)
    res_ref = mesolve(H, psi0, times, sc_ops, e_ops, args={"a":2})
    res = photocurrent_sesolve(H, psi0, times, sc_ops, e_ops, ntraj=ntraj,
                              nsubsteps=nsubsteps, store_measurement=True,
                              map_func=parallel_map, args={"a":2})

    assert_(all([np.mean(abs(res.expect[idx] - res_ref.expect[idx])) < tol
                 for idx in range(len(e_ops))]))

    assert_(len(res.measurement) == ntraj)
    assert_(all([m.shape == (len(times), len(sc_ops))
                 for m in res.measurement]))


def test_ssesolve_homodyne():
    "Stochastic: ssesolve: homodyne, time-dependent H"
    tol = 0.01

    N = 4
    gamma = 0.25
    ntraj = 25
    nsubsteps = 100
    a = destroy(N)

    H = [[a.dag() * a,f]]
    psi0 = coherent(N, 0.5)
    sc_ops = [np.sqrt(gamma) * a, np.sqrt(gamma) * a*0.5]
    e_ops = [a.dag() * a, a + a.dag(), (-1j)*(a - a.dag())]

    times = np.linspace(0, 2.5, 50)
    res_ref = mesolve(H, psi0, times, sc_ops, e_ops, args={"a":2})
    res = ssesolve(H, psi0, times, sc_ops, e_ops,
                   ntraj=ntraj, nsubsteps=nsubsteps,
                   method='homodyne', store_measurement=True,
                   map_func=parallel_map, args={"a":2})

    assert_(all([np.mean(abs(res.expect[idx] - res_ref.expect[idx])) < tol
                 for idx in range(len(e_ops))]))

    assert_(len(res.measurement) == ntraj)
    assert_(all([m.shape == (len(times), len(sc_ops))
                 for m in res.measurement]))


@pytest.mark.slow
def test_ssesolve_heterodyne():
    "Stochastic: ssesolve: heterodyne, time-dependent H"
    tol = 0.01

    N = 4
    gamma = 0.25
    ntraj = 25
    nsubsteps = 100
    a = destroy(N)

    H = [[a.dag() * a,f]]
    psi0 = coherent(N, 0.5)
    sc_ops = [np.sqrt(gamma) * a, np.sqrt(gamma) * a*0.5]
    e_ops = [a.dag() * a, a + a.dag(), (-1j)*(a - a.dag())]

    times = np.linspace(0, 2.5, 50)
    res_ref = mesolve(H, psi0, times, sc_ops, e_ops, args={"a":2})
    res = ssesolve(H, psi0, times, sc_ops, e_ops,
                   ntraj=ntraj, nsubsteps=nsubsteps,
                   method='heterodyne', store_measurement=True,
                   map_func=parallel_map, args={"a":2})

    assert_(all([np.mean(abs(res.expect[idx] - res_ref.expect[idx])) < tol
                 for idx in range(len(e_ops))]))

    assert_(len(res.measurement) == ntraj)
    assert_(all([m.shape == (len(times), len(sc_ops), 2)
                 for m in res.measurement]))


def f_dargs(a, args):
    return args["expect_op_3"] - 1


def test_ssesolve_feedback():
    "Stochastic: ssesolve: time-dependent H with feedback"
    tol = 0.01
    N = 4
    ntraj = 10
    nsubsteps = 100
    a = destroy(N)

    H = [num(N)]
    psi0 = coherent(N, 2.5)
    sc_ops = [[a + a.dag(), f_dargs]]
    e_ops = [a.dag() * a, a + a.dag(), (-1j)*(a - a.dag()), qeye(N)]

    times = np.linspace(0, 10, 101)
    res_ref = mesolve(H, psi0, times, sc_ops, e_ops,
                      args={"expect_op_3":qeye(N)})
    res = ssesolve(H, psi0, times, sc_ops, e_ops, solver=None, noise=1,
                   ntraj=ntraj, nsubsteps=nsubsteps, method='homodyne',
                   map_func=parallel_map, args={"expect_op_3":qeye(N)})


def test_ssesolve_bad_e_ops():
    tol = 0.01
    N = 4
    ntraj = 10
    nsubsteps = 100
    a = destroy(N)
    b = destroy(N-1)

    H = [num(N)]
    psi0 = coherent(N, 2.5)
    sc_ops = [a + a.dag()]
    e_ops = [a.dag() * a, a + a.dag(), (-1j)*(b - b.dag()), qeye(N+1)]
    times = np.linspace(0, 10, 101)
    with pytest.raises(TypeError) as exc:
        res = ssesolve(H, psi0, times, sc_ops, e_ops, solver=None, noise=1,
                       ntraj=ntraj, nsubsteps=nsubsteps, method='homodyne',
                       map_func=parallel_map)



import pytest
import numpy as np
import scipy.sparse
import qutip
from qutip.fastsparse import fast_csr_matrix
from qutip.cy.checks import (_test_sorting, _test_coo2csr_inplace_struct,
                             _test_csr2coo_struct, _test_coo2csr_struct)
from qutip.random_objects import rand_jacobi_rotation


def _unsorted_csr(N, density=0.5):
    M = scipy.sparse.diags(np.arange(N), 0, dtype=complex, format='csr')
    nvals = N**2 * density
    while M.nnz < 0.95*nvals:
        M = rand_jacobi_rotation(M)
    M = M.tocsr()
    return fast_csr_matrix((M.data, M.indices, M.indptr), shape=M.shape)


def sparse_arrays_equal(a, b):
    return not (a != b).data.any()


@pytest.mark.repeat(20)
def test_coo2csr_struct():
    "Cython structs : COO to CSR"
    A = qutip.rand_dm(5, 0.5).data
    assert sparse_arrays_equal(A, _test_coo2csr_struct(A.tocoo()))


@pytest.mark.repeat(20)
def test_indices_sort():
    "Cython structs : sort CSR indices inplace"
    A = _unsorted_csr(10, 0.25)
    B = A.copy()
    B.sort_indices()
    _test_sorting(A)
    assert np.all(A.data == B.data)
    assert np.all(A.indices == B.indices)


@pytest.mark.repeat(20)
def test_coo2csr_inplace_nosort():
    "Cython structs : COO to CSR inplace (no sort)"
    A = qutip.rand_dm(5, 0.5).data
    B = _test_coo2csr_inplace_struct(A.tocoo(), sorted=0)
    assert sparse_arrays_equal(A, B)


@pytest.mark.repeat(20)
def test_coo2csr_inplace_sort():
    "Cython structs : COO to CSR inplace (sorted)"
    A = qutip.rand_dm(5, 0.5).data
    B = _test_coo2csr_inplace_struct(A.tocoo(), sorted=1)
    assert sparse_arrays_equal(A, B)


@pytest.mark.repeat(20)
def test_csr2coo():
    "Cython structs : CSR to COO"
    A = qutip.rand_dm(5, 0.5).data
    B = A.tocoo()
    C = _test_csr2coo_struct(A)
    assert sparse_arrays_equal(B, C)



import numpy as np
import pytest
import qutip
import warnings


@pytest.mark.parametrize(['method', 'kwargs'], [
    pytest.param('direct', {}, id="direct"),
    pytest.param('direct', {'solver':'mkl'}, id="direct_mkl",
                 marks=pytest.mark.skipif(not qutip.settings.has_mkl,
                                          reason='MKL extensions not found.')),
    pytest.param('direct', {'return_info':True}, id="direct_info"),
    pytest.param('direct', {'sparse':False}, id="direct_dense"),
    pytest.param('direct', {'use_rcm':True}, id="direct_rcm"),
    pytest.param('direct', {'use_wbm':True}, id="direct_wbm"),
    pytest.param('eigen', {}, id="eigen"),
    pytest.param('eigen', {'use_rcm':True},  id="eigen_rcm"),
    pytest.param('svd', {}, id="svd"),
    pytest.param('power', {'mtol':1e-5}, id="power"),
    pytest.param('power', {'mtol':1e-5, 'solver':'mkl'}, id="power_mkl",
                 marks=pytest.mark.skipif(not qutip.settings.has_mkl,
                                          reason='MKL extensions not found.')),
    pytest.param('power-gmres', {'mtol':1e-1}, id="power-gmres"),
    pytest.param('power-gmres', {'mtol':1e-1, 'use_rcm':True, 'use_wbm':True},
                 id="power-gmres_perm"),
    pytest.param('power-bicgstab', {'use_precond':1}, id="power-bicgstab"),
    pytest.param('iterative-gmres', {}, id="iterative-gmres"),
    pytest.param('iterative-gmres', {'use_rcm':True, 'use_wbm':True},
                 id="iterative-gmres_perm"),
    pytest.param('iterative-bicgstab', {'return_info':True},
                 id="iterative-bicgstab"),
])
def test_qubit(method, kwargs):
    # thermal steadystate of a qubit: compare numerics with analytical formula
    sz = qutip.sigmaz()
    sm = qutip.destroy(2)

    H = 0.5 * 2 * np.pi * sz
    gamma1 = 0.05

    wth_vec = np.linspace(0.1, 3, 20)
    p_ss = np.zeros(np.shape(wth_vec))

    with warnings.catch_warnings():
        if 'use_wbm' in kwargs:
            # The deprecation has been fixed in dev.major
            warnings.simplefilter("ignore", category=DeprecationWarning)

        for idx, wth in enumerate(wth_vec):
            n_th = 1.0 / (np.exp(1.0 / wth) - 1)  # bath temperature
            c_op_list = []
            rate = gamma1 * (1 + n_th)
            c_op_list.append(np.sqrt(rate) * sm)
            rate = gamma1 * n_th
            c_op_list.append(np.sqrt(rate) * sm.dag())
            rho_ss = qutip.steadystate(H, c_op_list, method=method, **kwargs)
            if 'return_info' in kwargs:
                rho_ss, info = rho_ss
                assert isinstance(info, dict)
            p_ss[idx] = qutip.expect(sm.dag() * sm, rho_ss)

    p_ss_analytic = np.exp(-1.0 / wth_vec) / (1 + np.exp(-1.0 / wth_vec))
    np.testing.assert_allclose(p_ss_analytic, p_ss, atol=1e-5)


@pytest.mark.parametrize(['method', 'kwargs'], [
    pytest.param('direct', {}, id="direct"),
    pytest.param('direct', {'solver': 'mkl'}, id="direct_mkl",
                 marks=pytest.mark.skipif(not qutip.settings.has_mkl,
                                          reason='MKL extensions not found.')),
    pytest.param('direct', {'sparse': False}, id="direct_dense"),
    pytest.param('direct', {'use_rcm': True}, id="direct_rcm"),
    pytest.param('direct', {'use_wbm': True}, id="direct_wbm"),
    pytest.param('eigen', {}, id="eigen"),
    pytest.param('eigen', {'use_rcm': True},  id="eigen_rcm"),
    pytest.param('svd', {}, id="svd"),
])
def test_exact_solution_for_simple_methods(method, kwargs):
    # this tests that simple methods correctly determine the steadystate
    # with high accuracy for a small Liouvillian requiring correct weighting.
    H = qutip.identity(2)
    c_ops = [qutip.sigmam(), 1e-8 * qutip.sigmap()]
    rho_ss = qutip.steadystate(H, c_ops, method=method, **kwargs)
    expected_rho_ss = np.array([
        [1.e-16+0.j, 0.e+00-0.j],
        [0.e+00-0.j, 1.e+00+0.j],
    ])
    np.testing.assert_allclose(expected_rho_ss, rho_ss, atol=1e-16)
    assert rho_ss.tr() == pytest.approx(1, abs=1e-14)


@pytest.mark.parametrize(['method', 'kwargs'], [
    pytest.param('direct', {}, id="direct"),
    pytest.param('direct', {'sparse':False}, id="direct_dense"),
    pytest.param('eigen', {}, id="eigen"),
    pytest.param('power', {'mtol':1e-5}, id="power"),
    pytest.param('power-gmres', {'mtol':1e-1, 'use_precond':1}, id="power-gmres"),
    pytest.param('power-bicgstab', {'use_precond':1}, id="power-bicgstab"),
    pytest.param('iterative-lgmres', {'use_precond':1}, id="iterative-lgmres"),
    pytest.param('iterative-gmres', {}, id="iterative-gmres"),
    pytest.param('iterative-bicgstab', {}, id="iterative-bicgstab"),
])
def test_ho(method, kwargs):
    # thermal steadystate of an oscillator: compare numerics with analytical
    # formula
    a = qutip.destroy(35)
    H = 0.5 * 2 * np.pi * a.dag() * a
    gamma1 = 0.05

    wth_vec = np.linspace(0.1, 3, 20)
    p_ss = np.zeros(np.shape(wth_vec))

    for idx, wth in enumerate(wth_vec):

        n_th = 1.0 / (np.exp(1.0 / wth) - 1)  # bath temperature
        c_op_list = []
        rate = gamma1 * (1 + n_th)
        c_op_list.append(np.sqrt(rate) * a)
        rate = gamma1 * n_th
        c_op_list.append(np.sqrt(rate) * a.dag())
        rho_ss = qutip.steadystate(H, c_op_list, method=method, **kwargs)
        p_ss[idx] = np.real(qutip.expect(a.dag() * a, rho_ss))

    p_ss_analytic = 1.0 / (np.exp(1.0 / wth_vec) - 1)
    np.testing.assert_allclose(p_ss_analytic, p_ss, atol=1e-3)


@pytest.mark.parametrize(['method', 'kwargs'], [
    pytest.param('direct', {}, id="direct"),
    pytest.param('direct', {'sparse':False}, id="direct_dense"),
    pytest.param('eigen', {}, id="eigen"),
    pytest.param('svd', {}, id="svd"),
    pytest.param('power', {'mtol':1e-5}, id="power"),
    pytest.param('power-gmres', {'mtol':1e-1, 'use_precond':1, 'M':'iterative'},
                 id="power-gmres"),
    pytest.param('power-bicgstab', {'use_precond':1, 'M':'power'},
                 id="power-bicgstab"),
    pytest.param('iterative-gmres', {}, id="iterative-gmres"),
    pytest.param('iterative-bicgstab', {}, id="iterative-bicgstab"),
])
def test_driven_cavity(method, kwargs):
    N = 30
    Omega = 0.01 * 2 * np.pi
    Gamma = 0.05

    a = qutip.destroy(N)
    H = Omega * (a.dag() + a)
    c_ops = [np.sqrt(Gamma) * a]
    if 'use_precond' in kwargs:
        kwargs['M'] = qutip.build_preconditioner(H, c_ops, method=kwargs['M'])
    rho_ss = qutip.steadystate(H, c_ops, method=method, **kwargs)
    rho_ss_analytic = qutip.coherent_dm(N, -1.0j * (Omega)/(Gamma/2))

    np.testing.assert_allclose(rho_ss, rho_ss_analytic, atol=1e-4)
    assert rho_ss.tr() == pytest.approx(1, abs=1e-12)


@pytest.mark.parametrize(['method', 'kwargs'], [
    pytest.param('splu', {'sparse':False}, id="dense_direct"),
    pytest.param('numpy', {'sparse':False}, id="dense_numpy"),
    pytest.param('scipy', {'sparse':False}, id="dense_scipy"),
    pytest.param('splu', {}, id="splu"),
    pytest.param('spilu', {},  id="spilu"),
])
def test_pseudo_inverse(method, kwargs):
    N = 4
    a = qutip.destroy(N)
    H = (a.dag() + a)
    L = qutip.liouvillian(H, [a])
    rho = qutip.steadystate(L)
    Lpinv = qutip.pseudo_inverse(L, rho, method=method, **kwargs)
    np.testing.assert_allclose((L * Lpinv * L).full(), L.full())
    np.testing.assert_allclose((Lpinv * L * Lpinv).full(), Lpinv.full())
    assert rho.tr() == pytest.approx(1, abs=1e-15)


@pytest.mark.parametrize('sparse', [True, False])
def test_steadystate_floquet(sparse):
    """
    Test the steadystate solution for a periodically
    driven system.
    """
    N_c = 20

    a = qutip.destroy(N_c)
    a_d = a.dag()
    X_c = a + a_d

    w_c = 1

    A_l = 0.001
    w_l = w_c
    gam = 0.01

    H = w_c * a_d * a

    H_t = [H, [X_c, lambda t, args: args["A_l"] * np.cos(args["w_l"] * t)]]

    psi0 = qutip.fock(N_c, 0)

    args = {"A_l": A_l, "w_l": w_l}

    c_ops = []
    c_ops.append(np.sqrt(gam) * a)

    t_l = np.linspace(0, 20 / gam, 2000)

    expect_me = qutip.mesolve(H_t, psi0, t_l,
                        c_ops, [a_d * a], args=args).expect[0]

    rho_ss = qutip.steadystate_floquet(H, c_ops,
                                       A_l * X_c, w_l, n_it=3, sparse=sparse)
    expect_ss = qutip.expect(a_d * a, rho_ss)

    np.testing.assert_allclose(expect_me[-20:], expect_ss, atol=1e-3)
    assert rho_ss.tr() == pytest.approx(1, abs=1e-15)


def test_bad_options_steadystate():
    N = 4
    a = qutip.destroy(N)
    H = (a.dag() + a)
    c_ops = [a]
    with pytest.raises(ValueError):
        qutip.steadystate(H, c_ops, method='not a method')
    with pytest.raises(TypeError):
        qutip.steadystate(H, c_ops, method='direct', bad_opt=True)
    with pytest.raises(ValueError):
        qutip.steadystate(H, c_ops, method='direct', solver='Error')


def test_bad_options_pseudo_inverse():
    N = 4
    a = qutip.destroy(N)
    H = (a.dag() + a)
    L = qutip.liouvillian(H, [a])
    with pytest.raises(TypeError):
        qutip.pseudo_inverse(L, method='splu', bad_opt=True)
    with pytest.raises(ValueError):
        qutip.pseudo_inverse(L, method='not a method', sparse=False)
    with pytest.raises(ValueError):
        qutip.pseudo_inverse(L, method='not a method')


def test_bad_options_build_preconditioner():
    N = 4
    a = qutip.destroy(N)
    H = (a.dag() + a)
    c_ops = [a]
    with pytest.raises(TypeError):
        qutip.build_preconditioner(H, c_ops, method='power', bad_opt=True)
    with pytest.raises(ValueError):
        qutip.build_preconditioner(H, c_ops, method='not a method')


def test_bad_system():
    N = 4
    a = qutip.destroy(N)
    H = (a.dag() + a)
    with pytest.raises(TypeError):
        qutip.steadystate(H, [], method='direct')
    with pytest.raises(TypeError):
        qutip.steadystate(qutip.basis(N, N-1), [], method='direct')



import sys
import numpy as np
import scipy.sparse as sp
from ctypes import c_int, byref
from numpy.ctypeslib import ndpointer
import time
import qutip.settings as qset

# Load solver functions from mkl_lib
pardiso = qset.mkl_lib.pardiso
pardiso_delete = qset.mkl_lib.pardiso_handle_delete
if sys.maxsize > 2**32:  # Running 64-bit
    pardiso_64 = qset.mkl_lib.pardiso_64
    pardiso_delete_64 = qset.mkl_lib.pardiso_handle_delete_64


def _pardiso_parameters(hermitian, has_perm,
                        max_iter_refine,
                        scaling_vectors,
                        weighted_matching):
    iparm = np.zeros(64, dtype=np.int32)
    iparm[0] = 1  # Do not use default values
    iparm[1] = 3  # Use openmp nested dissection
    if has_perm:
        iparm[4] = 1
    iparm[7] = max_iter_refine  # Max number of iterative refinements
    if hermitian:
        iparm[9] = 8
    else:
        iparm[9] = 13
    if not hermitian:
        iparm[10] = int(scaling_vectors)
        iparm[12] = int(weighted_matching)  # Non-symmetric weighted matching
    iparm[17] = -1
    iparm[20] = 1
    iparm[23] = 1  # Parallel factorization
    iparm[26] = 0  # Check matrix structure
    iparm[34] = 1  # Use zero-based indexing
    return iparm


# Set error messages
pardiso_error_msgs = {
    '-1': 'Input inconsistant',
    '-2': 'Out of memory',
    '-3': 'Reordering problem',
    '-4':
        'Zero pivot, numerical factorization or iterative refinement problem',
    '-5': 'Unclassified internal error',
    '-6': 'Reordering failed',
    '-7': 'Diagonal matrix is singular',
    '-8': '32-bit integer overflow',
    '-9': 'Not enough memory for OOC',
    '-10': 'Error opening OOC files',
    '-11': 'Read/write error with OOC files',
    '-12': 'Pardiso-64 called from 32-bit library',
}


def _default_solver_args():
    return {
        'hermitian': False,
        'posdef': False,
        'max_iter_refine': 10,
        'scaling_vectors': True,
        'weighted_matching': True,
        'return_info': False,
    }


class mkl_lu:
    """
    Object pointing to LU factorization of a sparse matrix
    generated by mkl_splu.

    Methods
    -------
    solve(b, verbose=False)
        Solve system of equations using given RHS vector 'b'.
        Returns solution ndarray with same shape as input.

    info()
        Returns the statistics of the factorization and
        solution in the lu.info attribute.

    delete()
        Deletes the allocated solver memory.

    """
    def __init__(self, np_pt=None, dim=None, is_complex=None, data=None,
                 indptr=None, indices=None, iparm=None, np_iparm=None,
                 mtype=None, perm=None, np_perm=None, factor_time=None):
        self._np_pt = np_pt
        self._dim = dim
        self._is_complex = is_complex
        self._data = data
        self._indptr = indptr
        self._indices = indices
        self._iparm = iparm
        self._np_iparm = np_iparm
        self._mtype = mtype
        self._perm = perm
        self._np_perm = np_perm
        self._factor_time = factor_time
        self._solve_time = None

    def solve(self, b, verbose=None):
        b_shp = b.shape
        if b.ndim == 2 and b.shape[1] == 1:
            b = b.ravel()
            nrhs = 1
        elif b.ndim == 2 and b.shape[1] != 1:
            nrhs = b.shape[1]
            b = b.ravel(order='F')
        else:
            b = b.ravel()
            nrhs = 1

        data_type = np.complex128 if self._is_complex else np.float64
        if b.dtype != data_type:
            b = b.astype(np.complex128, copy=False)

        # Create solution array (x) and pointers to x and b
        x = np.zeros(b.shape, dtype=data_type, order='C')
        np_x = x.ctypes.data_as(ndpointer(data_type, ndim=1, flags='C'))
        np_b = b.ctypes.data_as(ndpointer(data_type, ndim=1, flags='C'))

        error = np.zeros(1, dtype=np.int32)
        np_error = error.ctypes.data_as(ndpointer(np.int32, ndim=1, flags='C'))

        # Call solver
        _solve_start = time.time()
        pardiso(
            self._np_pt,
            byref(c_int(1)),
            byref(c_int(1)),
            byref(c_int(self._mtype)),
            byref(c_int(33)),
            byref(c_int(self._dim)),
            self._data,
            self._indptr,
            self._indices,
            self._np_perm,
            byref(c_int(nrhs)),
            self._np_iparm,
            byref(c_int(0)),
            np_b,
            np_x,
            np_error,
        )
        self._solve_time = time.time() - _solve_start
        if error[0] != 0:
            raise Exception(pardiso_error_msgs[str(error[0])])

        if verbose:
            print('Solution Stage')
            print('--------------')
            print('Solution time:                  ',
                  round(self._solve_time, 4))
            print('Solution memory (Mb):           ',
                  round(self._iparm[16]/1024, 4))
            print('Number of iterative refinements:',
                  self._iparm[6])
            print('Total memory (Mb):              ',
                  round(sum(self._iparm[15:17])/1024, 4))
            print()
        return np.reshape(x, b_shp, order=('C' if nrhs == 1 else 'F'))

    def info(self):
        info = {'FactorTime': self._factor_time,
                'SolveTime': self._solve_time,
                'Factormem': round(self._iparm[15]/1024, 4),
                'Solvemem': round(self._iparm[16]/1024, 4),
                'IterRefine': self._iparm[6]}
        return info

    def delete(self):
        # Delete all data
        error = np.zeros(1, dtype=np.int32)
        np_error = error.ctypes.data_as(ndpointer(np.int32, ndim=1, flags='C'))
        pardiso(
            self._np_pt,
            byref(c_int(1)),
            byref(c_int(1)),
            byref(c_int(self._mtype)),
            byref(c_int(-1)),
            byref(c_int(self._dim)),
            self._data,
            self._indptr,
            self._indices,
            self._np_perm,
            byref(c_int(1)),
            self._np_iparm,
            byref(c_int(0)),
            byref(c_int(0)),
            byref(c_int(0)),
            np_error,
        )
        if error[0] == -10:
            raise Exception('Error freeing solver memory')


_MATRIX_TYPE_NAMES = {
    4: 'Complex Hermitian positive-definite',
    -4: 'Complex Hermitian indefinite',
    2: 'Real symmetric positive-definite',
    -2: 'Real symmetric indefinite',
    11: 'Real non-symmetric',
    13: 'Complex non-symmetric',
}


def _mkl_matrix_type(dtype, solver_args):
    if not solver_args['hermitian']:
        return 13 if dtype == np.complex128 else 11
    out = 4 if dtype == np.complex128 else 2
    return out if solver_args['posdef'] else -out


def mkl_splu(A, perm=None, verbose=False, **kwargs):
    """
    Returns the LU factorization of the sparse matrix A.

    Parameters
    ----------
    A : csr_matrix
        Sparse input matrix.
    perm : ndarray (optional)
        User defined matrix factorization permutation.
    verbose : bool {False, True}
        Report factorization details.

    Returns
    -------
    lu : mkl_lu
        Returns object containing LU factorization with a
        solve method for solving with a given RHS vector.

    """
    if not sp.isspmatrix_csr(A):
        raise TypeError('Input matrix must be in sparse CSR format.')

    if A.shape[0] != A.shape[1]:
        raise Exception('Input matrix must be square')

    dim = A.shape[0]
    solver_args = _default_solver_args()
    if set(kwargs) - set(solver_args):
        raise ValueError(
            "Unknown keyword arguments pass to mkl_splu: {!r}"
            .format(set(kwargs) - set(solver_args))
        )
    solver_args.update(kwargs)

    # If hermitian, then take upper-triangle of matrix only
    if solver_args['hermitian']:
        B = sp.triu(A, format='csr')
        A = B  # This gets around making a full copy of A in triu
    is_complex = bool(A.dtype == np.complex128)
    if not is_complex:
        A = sp.csr_matrix(A, dtype=np.float64, copy=False)
    data_type = A.dtype

    # Create pointer to internal memory
    pt = np.zeros(64, dtype=int)
    np_pt = pt.ctypes.data_as(ndpointer(int, ndim=1, flags='C'))

    # Create pointers to sparse matrix arrays
    data = A.data.ctypes.data_as(ndpointer(data_type, ndim=1, flags='C'))
    indptr = A.indptr.ctypes.data_as(ndpointer(np.int32, ndim=1, flags='C'))
    indices = A.indices.ctypes.data_as(ndpointer(np.int32, ndim=1, flags='C'))

    # Setup perm array
    if perm is None:
        perm = np.zeros(dim, dtype=np.int32)
        has_perm = 0
    else:
        has_perm = 1
    np_perm = perm.ctypes.data_as(ndpointer(np.int32, ndim=1, flags='C'))

    # setup iparm
    iparm = _pardiso_parameters(
        solver_args['hermitian'],
        has_perm,
        solver_args['max_iter_refine'],
        solver_args['scaling_vectors'],
        solver_args['weighted_matching'],
    )
    np_iparm = iparm.ctypes.data_as(ndpointer(np.int32, ndim=1, flags='C'))

    # setup call parameters
    mtype = _mkl_matrix_type(data_type, solver_args)

    if verbose:
        print('Solver Initialization')
        print('---------------------')
        print('Input matrix type: ', _MATRIX_TYPE_NAMES[mtype])
        print('Input matrix shape:', A.shape)
        print('Input matrix NNZ:  ', A.nnz)
        print()

    b = np.zeros(1, dtype=data_type)  # Input dummy RHS at this phase
    np_b = b.ctypes.data_as(ndpointer(data_type, ndim=1, flags='C'))
    x = np.zeros(1, dtype=data_type)  # Input dummy solution at this phase
    np_x = x.ctypes.data_as(ndpointer(data_type, ndim=1, flags='C'))

    error = np.zeros(1, dtype=np.int32)
    np_error = error.ctypes.data_as(ndpointer(np.int32, ndim=1, flags='C'))

    # Call solver
    _factor_start = time.time()
    pardiso(
        np_pt,
        byref(c_int(1)),
        byref(c_int(1)),
        byref(c_int(mtype)),
        byref(c_int(12)),
        byref(c_int(dim)),
        data,
        indptr,
        indices,
        np_perm,
        byref(c_int(1)),
        np_iparm,
        byref(c_int(0)),
        np_b,
        np_x,
        np_error,
    )
    _factor_time = time.time() - _factor_start
    if error[0] != 0:
        raise Exception(pardiso_error_msgs[str(error[0])])

    if verbose:
        print('Analysis and Factorization Stage')
        print('--------------------------------')
        print('Factorization time:       ', round(_factor_time, 4))
        print('Factorization memory (Mb):', round(iparm[15]/1024, 4))
        print('NNZ in LU factors:        ', iparm[17])
        print()

    return mkl_lu(np_pt, dim, is_complex, data, indptr, indices,
                  iparm, np_iparm, mtype, perm, np_perm, _factor_time)


def mkl_spsolve(A, b, perm=None, verbose=False, **kwargs):
    """
    Solves a sparse linear system of equations using the
    Intel MKL Pardiso solver.

    Parameters
    ----------
    A : csr_matrix
        Sparse matrix.
    b : ndarray or sparse matrix
        The vector or matrix representing the right hand side of the equation.
        If a vector, b.shape must be (n,) or (n, 1).
    perm : ndarray (optional)
        User defined matrix factorization permutation.

    Returns
    -------
    x : ndarray or csr_matrix
        The solution of the sparse linear equation.
        If b is a vector, then x is a vector of size A.shape[1]
        If b is a matrix, then x is a matrix of size (A.shape[1], b.shape[1])

    """
    lu = mkl_splu(A, perm=perm, verbose=verbose, **kwargs)
    b_is_sparse = sp.isspmatrix(b)
    b_shp = b.shape
    if b_is_sparse and b.shape[1] == 1:
        b = b.toarray()
        b_is_sparse = False
    elif b_is_sparse and b.shape[1] != 1:
        nrhs = b.shape[1]
        if lu._is_complex:
            b = sp.csc_matrix(b, dtype=np.complex128, copy=False)
        else:
            b = sp.csc_matrix(b, dtype=np.float64, copy=False)

    # Do dense RHS solving
    if not b_is_sparse:
        x = lu.solve(b, verbose=verbose)
    # Solve each RHS vec individually and convert to sparse
    else:
        data_segs = []
        row_segs = []
        col_segs = []
        for j in range(nrhs):
            bj = b[:, j].A.ravel()
            xj = lu.solve(bj)
            w = np.flatnonzero(xj)
            segment_length = w.shape[0]
            row_segs.append(w)
            col_segs.append(np.ones(segment_length, dtype=np.int32)*j)
            data_segs.append(np.asarray(xj[w], dtype=xj.dtype))
        sp_data = np.concatenate(data_segs)
        sp_row = np.concatenate(row_segs)
        sp_col = np.concatenate(col_segs)
        x = sp.csr_matrix((sp_data, (sp_row, sp_col)), shape=b_shp)

    info = lu.info()
    lu.delete()
    return (x, info) if kwargs.get('return_info', False) else x



__all__ = []

import numpy as np
import scipy.sparse as sp
from qutip.sparse import sp_reshape


def _ptrace(rho, sel):
    """
    Private function calculating the partial trace.
    """

    if isinstance(sel, int):
        sel = np.array([sel])
    else:
        sel = np.asarray(sel)

    if (sel < 0).any() or (sel >= len(rho.dims[0])).any():
        raise TypeError("Invalid selection index in ptrace.")

    drho = rho.dims[0]
    N = np.prod(drho)
    M = np.prod(np.asarray(drho).take(sel))

    if np.prod(rho.dims[1]) == 1:
        rho = rho * rho.dag()

    perm = sp.lil_matrix((M * M, N * N))
    # all elements in range(len(drho)) not in sel set
    rest = np.setdiff1d(np.arange(len(drho)), sel)
    ilistsel = _select(sel, drho)
    indsel = _list2ind(ilistsel, drho)
    ilistrest = _select(rest, drho)
    indrest = _list2ind(ilistrest, drho)
    irest = (indrest - 1) * N + indrest - 2
    # Possibly use parfor here if M > some value ?
    # We have to initialise like this to get a numpy array of lists rather than
    # a standard numpy 2d array.  Scipy >= 1.5 requires that we do this, rather
    # than just pass a 2d array, and scipy < 1.5 will accept it (because it's
    # actually the correct format).
    perm.rows = np.empty((M * M,), dtype=object)
    perm.data = np.empty((M * M,), dtype=object)
    for m in range(M * M):
        perm.rows[m] = list((irest
                             + (indsel[int(np.floor(m / M))] - 1)*N
                             + indsel[int(np.mod(m, M))]).T[0])
        perm.data[m] = [1.0] * len(perm.rows[m])
    perm = perm.tocsr()
    rhdata = perm * sp_reshape(rho.data, (np.prod(rho.shape), 1))
    rho1_data = sp_reshape(rhdata, (M, M))
    dims_kept0 = np.asarray(rho.dims[0]).take(sel)
    dims_kept1 = np.asarray(rho.dims[0]).take(sel)
    rho1_dims = [dims_kept0.tolist(), dims_kept1.tolist()]
    rho1_shape = [np.prod(dims_kept0), np.prod(dims_kept1)]
    return rho1_data, rho1_dims, rho1_shape


def _list2ind(ilist, dims):
    """!
    Private function returning indicies
    """
    ilist = np.asarray(ilist)
    dims = np.asarray(dims)
    irev = np.fliplr(ilist) - 1
    fact = np.append(np.array([1]), (np.cumprod(np.flipud(dims)[:-1])))
    fact = fact.reshape(len(fact), 1)
    return np.array(np.sort(np.dot(irev, fact) + 1, 0), dtype=int)


def _select(sel, dims):
    """
    Private function finding selected components
    """
    sel = np.asarray(sel)  # make sure sel is np.array
    dims = np.asarray(dims)  # make sure dims is np.array
    rlst = dims.take(sel)
    rprod = np.prod(rlst)
    ilist = np.ones((rprod, len(dims)), dtype=int)
    counter = np.arange(rprod)
    for k in range(len(sel)):
        ilist[:, sel[k]] = np.remainder(
            np.fix(counter / np.prod(dims[sel[k + 1:]])), dims[sel[k]]) + 1
    return ilist



#cython: language_level=3

cimport numpy as cnp

from qutip.cy.brtools cimport spec_func


cdef void cop_super_mult_openmp(complex[::1,:] cop, complex[::1,:] evecs,  double complex * vec,
                    double complex alpha,
                    double complex * out,
                    unsigned int nrows,
                    unsigned int omp_thresh,
                    unsigned int nthr,
                    double atol) except *


cdef void br_term_mult_openmp(double t, complex[::1,:] A, complex[::1,:] evecs,
                double[:,::1] skew, double dw_min, spec_func spectral,
                double complex * vec, double complex * out,
                unsigned int nrows, int use_secular,
                double sec_cutoff,
                unsigned int omp_thresh,
                unsigned int nthr,
                double atol) except *



#cython: language_level=3

cdef complex _spline_complex_t_second(double x, double[::1] t,
                                      complex[::1] y, complex[::1] M,
                                      int N)

cdef complex _spline_complex_cte_second(double x, double[::1] t,
                                        complex[::1] y, complex[::1] M,
                                        int N, double dt)

cdef double _spline_float_t_second(double x, double[::1] t,
                                   double[::1] y, double[::1] M,
                                   int N)

cdef double _spline_float_cte_second(double x, double[::1] t,
                                     double[::1] y, double[::1] M,
                                     int N, double dt)

cdef double _step_float_cte(double x, double[::1] t, double[::1] y, int n_t)

cdef complex _step_complex_cte(double x, double[::1] t, complex[::1] y, int n_t)

cdef double _step_float_t(double x, double[::1] t, double[::1] y, int n_t)

cdef complex _step_complex_t(double x, double[::1] t, complex[::1] y, int n_t)



import os

def _cython_build_cleanup(tdname, build_dir=None):
    if build_dir is None:
        build_dir = os.path.join(os.path.expanduser('~'), '.pyxbld')
    
    # Remove tdname.pyx
    pyx_file = tdname + ".pyx"
    try:
        os.remove(pyx_file)
    except:
        pass
    
    # Remove temp build files
    for dirpath, subdirs, files in os.walk(build_dir):
        for f in files:
            if f.startswith(tdname):
                try:
                    os.remove(os.path.join(dirpath,f))
                except:
                    pass



__all__ = ['Bloch']

import os

from numpy import (ndarray, array, linspace, pi, outer, cos, sin, ones, size,
                   sqrt, real, mod, append, ceil, arange)
import numpy as np

from packaging.version import parse as parse_version

from qutip.qobj import Qobj
from qutip.expect import expect
from qutip.operators import sigmax, sigmay, sigmaz

try:
    import matplotlib
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib.patches import FancyArrowPatch
    from mpl_toolkits.mplot3d import proj3d

    # Define a custom _axes3D function based on the matplotlib version.
    # The auto_add_to_figure keyword is new for matplotlib>=3.4.
    if parse_version(matplotlib.__version__) >= parse_version('3.4'):
        def _axes3D(fig, *args, **kwargs):
            ax = Axes3D(fig, *args, auto_add_to_figure=False, **kwargs)
            return fig.add_axes(ax)
    else:
        def _axes3D(*args, **kwargs):
            return Axes3D(*args, **kwargs)

    class Arrow3D(FancyArrowPatch):
        def __init__(self, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)

            self._verts3d = xs, ys, zs

        def draw(self, renderer):
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)

            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            FancyArrowPatch.draw(self, renderer)

        def do_3d_projection(self, renderer=None):
            # only called by matplotlib >= 3.5
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            return np.min(zs)
except ImportError:
    pass

try:
    from IPython.display import display
except ImportError:
    pass


class Bloch:
    r"""
    Class for plotting data on the Bloch sphere.  Valid data can be either
    points, vectors, or Qobj objects.

    Attributes
    ----------
    axes : matplotlib.axes.Axes
        User supplied Matplotlib axes for Bloch sphere animation.
    fig : matplotlib.figure.Figure
        User supplied Matplotlib Figure instance for plotting Bloch sphere.
    font_color : str, default 'black'
        Color of font used for Bloch sphere labels.
    font_size : int, default 20
        Size of font used for Bloch sphere labels.
    frame_alpha : float, default 0.1
        Sets transparency of Bloch sphere frame.
    frame_color : str, default 'gray'
        Color of sphere wireframe.
    frame_width : int, default 1
        Width of wireframe.
    point_color : list, default ["b", "r", "g", "#CC6600"]
        List of colors for Bloch sphere point markers to cycle through, i.e.
        by default, points 0 and 4 will both be blue ('b').
    point_marker : list, default ["o", "s", "d", "^"]
        List of point marker shapes to cycle through.
    point_size : list, default [25, 32, 35, 45]
        List of point marker sizes. Note, not all point markers look the same
        size when plotted!
    sphere_alpha : float, default 0.2
        Transparency of Bloch sphere itself.
    sphere_color : str, default '#FFDDDD'
        Color of Bloch sphere.
    figsize : list, default [7, 7]
        Figure size of Bloch sphere plot.  Best to have both numbers the same;
        otherwise you will have a Bloch sphere that looks like a football.
    vector_color : list, ["g", "#CC6600", "b", "r"]
        List of vector colors to cycle through.
    vector_width : int, default 5
        Width of displayed vectors.
    vector_style : str, default '-\|>'
        Vector arrowhead style (from matplotlib's arrow style).
    vector_mutation : int, default 20
        Width of vectors arrowhead.
    view : list, default [-60, 30]
        Azimuthal and Elevation viewing angles.
    xlabel : list, default ["$x$", ""]
        List of strings corresponding to +x and -x axes labels, respectively.
    xlpos : list, default [1.1, -1.1]
        Positions of +x and -x labels respectively.
    ylabel : list, default ["$y$", ""]
        List of strings corresponding to +y and -y axes labels, respectively.
    ylpos : list, default [1.2, -1.2]
        Positions of +y and -y labels respectively.
    zlabel : list, default ['$\\left\|0\\right>$', '$\\left\|1\\right>$']
        List of strings corresponding to +z and -z axes labels, respectively.
    zlpos : list, default [1.2, -1.2]
        Positions of +z and -z labels respectively.
    """
    def __init__(self, fig=None, axes=None, view=None, figsize=None,
                 background=False):
        # Figure and axes
        self.fig = fig
        self._ext_fig = fig is not None
        self.axes = axes
        # Background axes, default = False
        self.background = background
        # The size of the figure in inches, default = [5,5].
        self.figsize = figsize if figsize else [5, 5]
        # Azimuthal and Elvation viewing angles, default = [-60,30].
        self.view = view if view else [-60, 30]
        # Color of Bloch sphere, default = #FFDDDD
        self.sphere_color = '#FFDDDD'
        # Transparency of Bloch sphere, default = 0.2
        self.sphere_alpha = 0.2
        # Color of wireframe, default = 'gray'
        self.frame_color = 'gray'
        # Width of wireframe, default = 1
        self.frame_width = 1
        # Transparency of wireframe, default = 0.2
        self.frame_alpha = 0.2
        # Labels for x-axis (in LaTex), default = ['$x$', '']
        self.xlabel = ['$x$', '']
        # Position of x-axis labels, default = [1.2, -1.2]
        self.xlpos = [1.2, -1.2]
        # Labels for y-axis (in LaTex), default = ['$y$', '']
        self.ylabel = ['$y$', '']
        # Position of y-axis labels, default = [1.1, -1.1]
        self.ylpos = [1.2, -1.2]
        # Labels for z-axis (in LaTex),
        # default = [r'$\left\|0\right>$', r'$\left|1\right>$']
        self.zlabel = [r'$\left|0\right>$', r'$\left|1\right>$']
        # Position of z-axis labels, default = [1.2, -1.2]
        self.zlpos = [1.2, -1.2]
        # ---font options---
        # Color of fonts, default = 'black'
        self.font_color = 'black'
        # Size of fonts, default = 20
        self.font_size = 20

        # ---vector options---
        # List of colors for Bloch vectors, default = ['b','g','r','y']
        self.vector_color = ['g', '#CC6600', 'b', 'r']
        #: Width of Bloch vectors, default = 5
        self.vector_width = 3
        #: Style of Bloch vectors, default = '-\|>' (or 'simple')
        self.vector_style = '-|>'
        #: Sets the width of the vectors arrowhead
        self.vector_mutation = 20

        # ---point options---
        # List of colors for Bloch point markers, default = ['b','g','r','y']
        self.point_color = ['b', 'r', 'g', '#CC6600']
        # Size of point markers, default = 25
        self.point_size = [25, 32, 35, 45]
        # Shape of point markers, default = ['o','^','d','s']
        self.point_marker = ['o', 's', 'd', '^']

        # ---data lists---
        # Data for point markers
        self.points = []
        # Data for Bloch vectors
        self.vectors = []
        # Transparency of vectors, alpha value from 0 to 1
        self.vector_alpha = []
        # Data for annotations
        self.annotations = []
        # Number of times sphere has been saved
        self.savenum = 0
        # Style of points, 'm' for multiple colors, 's' for single color
        self.point_style = []
        # Transparency of points, alpha value from 0 to 1
        self.point_alpha = []
        # Data for line segment
        self._lines = []
        # Data for arcs and arc style
        self._arcs = []

    def set_label_convention(self, convention):
        """Set x, y and z labels according to one of conventions.

        Parameters
        ----------
        convention : string
            One of the following:

            - "original"
            - "xyz"
            - "sx sy sz"
            - "01"
            - "polarization jones"
            - "polarization jones letters"
              see also: https://en.wikipedia.org/wiki/Jones_calculus
            - "polarization stokes"
              see also: https://en.wikipedia.org/wiki/Stokes_parameters
        """
        ketex = "$\\left.|%s\\right\\rangle$"
        # \left.| is on purpose, so that every ket has the same size

        if convention == "original":
            self.xlabel = ['$x$', '']
            self.ylabel = ['$y$', '']
            self.zlabel = ['$\\left|0\\right>$', '$\\left|1\\right>$']
        elif convention == "xyz":
            self.xlabel = ['$x$', '']
            self.ylabel = ['$y$', '']
            self.zlabel = ['$z$', '']
        elif convention == "sx sy sz":
            self.xlabel = ['$s_x$', '']
            self.ylabel = ['$s_y$', '']
            self.zlabel = ['$s_z$', '']
        elif convention == "01":
            self.xlabel = ['', '']
            self.ylabel = ['', '']
            self.zlabel = ['$\\left|0\\right>$', '$\\left|1\\right>$']
        elif convention == "polarization jones":
            self.xlabel = [ketex % "\\nearrow\\hspace{-1.46}\\swarrow",
                           ketex % "\\nwarrow\\hspace{-1.46}\\searrow"]
            self.ylabel = [ketex % "\\circlearrowleft", ketex %
                           "\\circlearrowright"]
            self.zlabel = [ketex % "\\leftrightarrow", ketex % "\\updownarrow"]
        elif convention == "polarization jones letters":
            self.xlabel = [ketex % "D", ketex % "A"]
            self.ylabel = [ketex % "L", ketex % "R"]
            self.zlabel = [ketex % "H", ketex % "V"]
        elif convention == "polarization stokes":
            self.ylabel = ["$\\nearrow\\hspace{-1.46}\\swarrow$",
                           "$\\nwarrow\\hspace{-1.46}\\searrow$"]
            self.zlabel = ["$\\circlearrowleft$", "$\\circlearrowright$"]
            self.xlabel = ["$\\leftrightarrow$", "$\\updownarrow$"]
        else:
            raise Exception("No such convention.")

    def __str__(self):
        s = ""
        s += "Bloch data:\n"
        s += "-----------\n"
        s += "Number of points:  " + str(len(self.points)) + "\n"
        s += "Number of vectors: " + str(len(self.vectors)) + "\n"
        s += "\n"
        s += "Bloch sphere properties:\n"
        s += "------------------------\n"
        s += "font_color:      " + str(self.font_color) + "\n"
        s += "font_size:       " + str(self.font_size) + "\n"
        s += "frame_alpha:     " + str(self.frame_alpha) + "\n"
        s += "frame_color:     " + str(self.frame_color) + "\n"
        s += "frame_width:     " + str(self.frame_width) + "\n"
        s += "point_color:     " + str(self.point_color) + "\n"
        s += "point_marker:    " + str(self.point_marker) + "\n"
        s += "point_size:      " + str(self.point_size) + "\n"
        s += "sphere_alpha:    " + str(self.sphere_alpha) + "\n"
        s += "sphere_color:    " + str(self.sphere_color) + "\n"
        s += "figsize:         " + str(self.figsize) + "\n"
        s += "vector_color:    " + str(self.vector_color) + "\n"
        s += "vector_width:    " + str(self.vector_width) + "\n"
        s += "vector_style:    " + str(self.vector_style) + "\n"
        s += "vector_mutation: " + str(self.vector_mutation) + "\n"
        s += "view:            " + str(self.view) + "\n"
        s += "xlabel:          " + str(self.xlabel) + "\n"
        s += "xlpos:           " + str(self.xlpos) + "\n"
        s += "ylabel:          " + str(self.ylabel) + "\n"
        s += "ylpos:           " + str(self.ylpos) + "\n"
        s += "zlabel:          " + str(self.zlabel) + "\n"
        s += "zlpos:           " + str(self.zlpos) + "\n"
        return s

    def _repr_png_(self):
        from IPython.core.pylabtools import print_figure
        self.render()
        fig_data = print_figure(self.fig, 'png')
        plt.close(self.fig)
        return fig_data

    def _repr_svg_(self):
        from IPython.core.pylabtools import print_figure
        self.render()
        fig_data = print_figure(self.fig, 'svg').decode('utf-8')
        plt.close(self.fig)
        return fig_data

    def clear(self):
        """Resets Bloch sphere data sets to empty.
        """
        self.points = []
        self.vectors = []
        self.point_style = []
        self.point_alpha = []
        self.vector_alpha = []
        self.annotations = []
        self._lines = []
        self._arcs = []

    def add_points(self, points, meth='s', alpha=1.0):
        """Add a list of data points to bloch sphere.

        Parameters
        ----------
        points : array_like
            Collection of data points.

        meth : {'s', 'm', 'l'}
            Type of points to plot, use 'm' for multicolored, 'l' for points
            connected with a line.

        alpha : float, default=1.
            Transparency value for the vectors. Values between 0 and 1.

        .. note::

           When using ``meth=l`` in QuTiP 4.6, the line transparency defaulted
           to ``0.75`` and there was no way to alter it.
           When the ``alpha`` parameter was added in QuTiP 4.7, the default
           became ``alpha=1.0`` for values of ``meth``.
        """
        if not isinstance(points[0], (list, tuple, ndarray)):
            points = [[points[0]], [points[1]], [points[2]]]
        points = array(points)
        if meth == 's':
            if len(points[0]) == 1:
                pnts = array([[points[0][0]], [points[1][0]], [points[2][0]]])
                pnts = append(pnts, points, axis=1)
            else:
                pnts = points
            self.points.append(pnts)
            self.point_style.append('s')
            self.point_alpha.append(alpha)
        elif meth == 'l':
            self.points.append(points)
            self.point_style.append('l')
            self.point_alpha.append(alpha)
        else:
            self.points.append(points)
            self.point_style.append('m')
            self.point_alpha.append(alpha)

    def add_states(self, state, kind='vector', alpha=1.0):
        """Add a state vector Qobj to Bloch sphere.

        Parameters
        ----------
        state : Qobj
            Input state vector.

        kind : {'vector', 'point'}
            Type of object to plot.

        alpha : float, default=1.
            Transparency value for the vectors. Values between 0 and 1.
        """
        if isinstance(state, Qobj):
            state = [state]

        for st in state:
            vec = [expect(sigmax(), st),
                   expect(sigmay(), st),
                   expect(sigmaz(), st)]

            if kind == 'vector':
                self.add_vectors(vec, alpha=alpha)
            elif kind == 'point':
                self.add_points(vec, alpha=alpha)

    def add_vectors(self, vectors, alpha=1.0):
        """Add a list of vectors to Bloch sphere.

        Parameters
        ----------
        vectors : array_like
            Array with vectors of unit length or smaller.

        alpha : float, default=1.
            Transparency value for the vectors. Values between 0 and 1.
        """
        if isinstance(vectors[0], (list, tuple, ndarray)):
            for vec in vectors:
                self.vectors.append(vec)
                self.vector_alpha.append(alpha)
        else:
            self.vectors.append(vectors)
            self.vector_alpha.append(alpha)

    def add_annotation(self, state_or_vector, text, **kwargs):
        """
        Add a text or LaTeX annotation to Bloch sphere, parametrized by a qubit
        state or a vector.

        Parameters
        ----------
        state_or_vector : Qobj/array/list/tuple
            Position for the annotaion.
            Qobj of a qubit or a vector of 3 elements.

        text : str
            Annotation text.
            You can use LaTeX, but remember to use raw string
            e.g. r"$\\langle x \\rangle$"
            or escape backslashes
            e.g. "$\\\\langle x \\\\rangle$".

        kwargs :
            Options as for mplot3d.axes3d.text, including:
            fontsize, color, horizontalalignment, verticalalignment.

        """
        if isinstance(state_or_vector, Qobj):
            vec = [expect(sigmax(), state_or_vector),
                   expect(sigmay(), state_or_vector),
                   expect(sigmaz(), state_or_vector)]
        elif isinstance(state_or_vector, (list, ndarray, tuple)) \
                and len(state_or_vector) == 3:
            vec = state_or_vector
        else:
            raise Exception("Position needs to be specified by a qubit " +
                            "state or a 3D vector.")
        self.annotations.append({'position': vec,
                                 'text': text,
                                 'opts': kwargs})

    def add_arc(self, start, end, fmt="b", steps=None, **kwargs):
        """Adds an arc between two points on a sphere. The arc is set to be
        blue solid curve by default.

        The start and end points must be on the same sphere (i.e. have the
        same radius) but need not be on the unit sphere.

        Parameters
        ----------
        start : Qobj or array-like
            Array with cartesian coordinates of the first point, or a state
            vector or density matrix that can be mapped to a point on or
            within the Bloch sphere.
        end : Qobj or array-like
            Array with cartesian coordinates of the second point, or a state
            vector or density matrix that can be mapped to a point on or
            within the Bloch sphere.
        fmt : str, default: "b"
            A matplotlib format string for rendering the arc.
        steps : int, default: None
            The number of segments to use when rendering the arc. The default
            uses 100 steps times the distance between the start and end points,
            with a minimum of 2 steps.
        **kwargs : dict
            Additional parameters to pass to the matplotlib .plot function
            when rendering this arc.
        """
        if isinstance(start, Qobj):
            pt1 = [
                expect(sigmax(), start),
                expect(sigmay(), start),
                expect(sigmaz(), start),
            ]
        else:
            pt1 = start

        if isinstance(end, Qobj):
            pt2 = [
                expect(sigmax(), end),
                expect(sigmay(), end),
                expect(sigmaz(), end),
            ]
        else:
            pt2 = end

        pt1 = np.asarray(pt1)
        pt2 = np.asarray(pt2)

        len1 = np.linalg.norm(pt1)
        len2 = np.linalg.norm(pt2)
        if len1 < 1e-12 or len2 < 1e-12:
            raise ValueError('Polar and azimuthal angles undefined at origin.')
        elif abs(len1 - len2) > 1e-12:
            raise ValueError("Points not on the same sphere.")
        elif (pt1 == pt2).all():
            raise ValueError(
                "Start and end represent the same point. No arc can be formed."
            )
        elif (pt1 == -pt2).all():
            raise ValueError(
                "Start and end are diagonally opposite, no unique arc is"
                " possible."
            )

        if steps is None:
            steps = int(np.linalg.norm(pt1 - pt2) * 100)
            steps = max(2, steps)
        t = np.linspace(0, 1, steps)
        # All the points in this line are contained in the plane defined
        # by pt1, pt2 and the origin.
        line = pt1[:, np.newaxis] * t + pt2[:, np.newaxis] * (1 - t)
        # Normalize all the points in the line so that are distance len1 from
        # the origin.
        arc = line * len1 / np.linalg.norm(line, axis=0)
        self._arcs.append([arc, fmt, kwargs])

    def add_line(self, start, end, fmt="k", **kwargs):
        """Adds a line segment connecting two points on the bloch sphere.

        The line segment is set to be a black solid line by default.

        Parameters
        ----------
        start : Qobj or array-like
            Array with cartesian coordinates of the first point, or a state
            vector or density matrix that can be mapped to a point on or
            within the Bloch sphere.
        end : Qobj or array-like
            Array with cartesian coordinates of the second point, or a state
            vector or density matrix that can be mapped to a point on or
            within the Bloch sphere.
        fmt : str, default: "k"
            A matplotlib format string for rendering the line.
        **kwargs : dict
            Additional parameters to pass to the matplotlib .plot function
            when rendering this line.
        """
        if isinstance(start, Qobj):
            pt1 = [
                expect(sigmax(), start),
                expect(sigmay(), start),
                expect(sigmaz(), start),
            ]
        else:
            pt1 = start

        if isinstance(end, Qobj):
            pt2 = [
                expect(sigmax(), end),
                expect(sigmay(), end),
                expect(sigmaz(), end),
            ]
        else:
            pt2 = end

        pt1 = np.asarray(pt1)
        pt2 = np.asarray(pt2)

        x = [pt1[1], pt2[1]]
        y = [-pt1[0], -pt2[0]]
        z = [pt1[2], pt2[2]]
        v = [x, y, z]
        self._lines.append([v, fmt, kwargs])

    def make_sphere(self):
        """
        Plots Bloch sphere and data sets.
        """
        self.render()

    def run_from_ipython(self):
        try:
            __IPYTHON__
            return True
        except NameError:
            return False

    def _is_inline_backend(self):
        backend = matplotlib.get_backend()
        return backend == "module://matplotlib_inline.backend_inline"

    def render(self):
        """
        Render the Bloch sphere and its data sets in on given figure and axes.
        """
        if not self._ext_fig and not self._is_inline_backend():
            # If no external figure was supplied, we check to see if the
            # figure we created in a previous call to .render() has been
            # closed, and re-create if has been. This has the unfortunate
            # side effect of losing any modifications made to the axes or
            # figure, but the alternative is to crash the matplotlib backend.
            #
            # The inline backend used by, e.g. jupyter notebooks, is happy to
            # use closed figures so we leave those figures intact.
            if (
                self.fig is not None and
                not plt.fignum_exists(self.fig.number)
            ):
                self.fig = None
                self.axes = None

        if self.fig is None:
            self.fig = plt.figure(figsize=self.figsize)
            if self._is_inline_backend():
                # We immediately close the inline figure do avoid displaying
                # the figure twice when .show() calls display.
                plt.close(self.fig)

        if self.axes is None:
            self.axes = _axes3D(self.fig, azim=self.view[0], elev=self.view[1])

        # Clearing the axes is horrifically slow and loses a lot of the
        # axes state, but matplotlib doesn't seem to provide a better way
        # to redraw Axes3D. :/
        self.axes.clear()
        self.axes.grid(False)
        if self.background:
            self.axes.set_xlim3d(-1.3, 1.3)
            self.axes.set_ylim3d(-1.3, 1.3)
            self.axes.set_zlim3d(-1.3, 1.3)
        else:
            self.axes.set_axis_off()
            self.axes.set_xlim3d(-0.7, 0.7)
            self.axes.set_ylim3d(-0.7, 0.7)
            self.axes.set_zlim3d(-0.7, 0.7)
        # Manually set aspect ratio to fit a square bounding box.
        # Matplotlib did this stretching for < 3.3.0, but not above.
        if parse_version(matplotlib.__version__) >= parse_version('3.3'):
            self.axes.set_box_aspect((1, 1, 1))
        if not self.background:
            self.plot_axes()

        self.plot_back()
        self.plot_points()
        self.plot_vectors()
        self.plot_lines()
        self.plot_arcs()
        self.plot_front()
        self.plot_axes_labels()
        self.plot_annotations()
        # Trigger an update of the Bloch sphere if it is already shown:
        self.fig.canvas.draw()

    def plot_back(self):
        # back half of sphere
        u = linspace(0, pi, 25)
        v = linspace(0, pi, 25)
        x = outer(cos(u), sin(v))
        y = outer(sin(u), sin(v))
        z = outer(ones(size(u)), cos(v))
        self.axes.plot_surface(x, y, z, rstride=2, cstride=2,
                               color=self.sphere_color, linewidth=0,
                               alpha=self.sphere_alpha)
        # wireframe
        self.axes.plot_wireframe(x, y, z, rstride=5, cstride=5,
                                 color=self.frame_color,
                                 alpha=self.frame_alpha)
        # equator
        self.axes.plot(1.0 * cos(u), 1.0 * sin(u), zs=0, zdir='z',
                       lw=self.frame_width, color=self.frame_color)
        self.axes.plot(1.0 * cos(u), 1.0 * sin(u), zs=0, zdir='x',
                       lw=self.frame_width, color=self.frame_color)

    def plot_front(self):
        # front half of sphere
        u = linspace(-pi, 0, 25)
        v = linspace(0, pi, 25)
        x = outer(cos(u), sin(v))
        y = outer(sin(u), sin(v))
        z = outer(ones(size(u)), cos(v))
        self.axes.plot_surface(x, y, z, rstride=2, cstride=2,
                               color=self.sphere_color, linewidth=0,
                               alpha=self.sphere_alpha)
        # wireframe
        self.axes.plot_wireframe(x, y, z, rstride=5, cstride=5,
                                 color=self.frame_color,
                                 alpha=self.frame_alpha)
        # equator
        self.axes.plot(1.0 * cos(u), 1.0 * sin(u),
                       zs=0, zdir='z', lw=self.frame_width,
                       color=self.frame_color)
        self.axes.plot(1.0 * cos(u), 1.0 * sin(u),
                       zs=0, zdir='x', lw=self.frame_width,
                       color=self.frame_color)

    def plot_axes(self):
        # axes
        span = linspace(-1.0, 1.0, 2)
        self.axes.plot(span, 0 * span, zs=0, zdir='z', label='X',
                       lw=self.frame_width, color=self.frame_color)
        self.axes.plot(0 * span, span, zs=0, zdir='z', label='Y',
                       lw=self.frame_width, color=self.frame_color)
        self.axes.plot(0 * span, span, zs=0, zdir='y', label='Z',
                       lw=self.frame_width, color=self.frame_color)

    def plot_axes_labels(self):
        # axes labels
        opts = {'fontsize': self.font_size,
                'color': self.font_color,
                'horizontalalignment': 'center',
                'verticalalignment': 'center'}
        self.axes.text(0, -self.xlpos[0], 0, self.xlabel[0], **opts)
        self.axes.text(0, -self.xlpos[1], 0, self.xlabel[1], **opts)

        self.axes.text(self.ylpos[0], 0, 0, self.ylabel[0], **opts)
        self.axes.text(self.ylpos[1], 0, 0, self.ylabel[1], **opts)

        self.axes.text(0, 0, self.zlpos[0], self.zlabel[0], **opts)
        self.axes.text(0, 0, self.zlpos[1], self.zlabel[1], **opts)

        for a in (self.axes.xaxis.get_ticklines() +
                  self.axes.xaxis.get_ticklabels()):
            a.set_visible(False)
        for a in (self.axes.yaxis.get_ticklines() +
                  self.axes.yaxis.get_ticklabels()):
            a.set_visible(False)
        for a in (self.axes.zaxis.get_ticklines() +
                  self.axes.zaxis.get_ticklabels()):
            a.set_visible(False)

    def plot_vectors(self):
        # -X and Y data are switched for plotting purposes
        for k in range(len(self.vectors)):

            xs3d = self.vectors[k][1] * array([0, 1])
            ys3d = -self.vectors[k][0] * array([0, 1])
            zs3d = self.vectors[k][2] * array([0, 1])

            color = self.vector_color[mod(k, len(self.vector_color))]
            alpha = self.vector_alpha[k]

            if self.vector_style == '':
                # simple line style
                self.axes.plot(xs3d, ys3d, zs3d,
                               zs=0, zdir='z', label='Z',
                               lw=self.vector_width, color=color,
                               alpha=alpha)
            else:
                # decorated style, with arrow heads
                a = Arrow3D(xs3d, ys3d, zs3d,
                            mutation_scale=self.vector_mutation,
                            lw=self.vector_width,
                            arrowstyle=self.vector_style,
                            color=color, alpha=alpha)

                self.axes.add_artist(a)

    def plot_points(self):
        # -X and Y data are switched for plotting purposes
        for k in range(len(self.points)):
            num = len(self.points[k][0])
            dist = [sqrt(self.points[k][0][j] ** 2 +
                         self.points[k][1][j] ** 2 +
                         self.points[k][2][j] ** 2) for j in range(num)]
            if any(abs(dist - dist[0]) / dist[0] > 1e-12):
                # combine arrays so that they can be sorted together
                zipped = list(zip(dist, range(num)))
                zipped.sort()  # sort rates from lowest to highest
                dist, indperm = zip(*zipped)
                indperm = array(indperm)
            else:
                indperm = arange(num)
            if self.point_style[k] == 's':
                self.axes.scatter(
                    real(self.points[k][1][indperm]),
                    - real(self.points[k][0][indperm]),
                    real(self.points[k][2][indperm]),
                    s=self.point_size[mod(k, len(self.point_size))],
                    alpha=self.point_alpha[k],
                    edgecolor=None,
                    zdir='z',
                    color=self.point_color[mod(k, len(self.point_color))],
                    marker=self.point_marker[mod(k, len(self.point_marker))])

            elif self.point_style[k] == 'm':
                pnt_colors = array(self.point_color *
                                   int(ceil(num / float(
                                       len(self.point_color)))))

                pnt_colors = pnt_colors[0:num]
                pnt_colors = list(pnt_colors[indperm])
                marker = self.point_marker[mod(k, len(self.point_marker))]
                s = self.point_size[mod(k, len(self.point_size))]
                self.axes.scatter(real(self.points[k][1][indperm]),
                                  -real(self.points[k][0][indperm]),
                                  real(self.points[k][2][indperm]),
                                  s=s, alpha=self.point_alpha[k],
                                  edgecolor=None, zdir='z',
                                  color=pnt_colors, marker=marker)

            elif self.point_style[k] == 'l':
                color = self.point_color[mod(k, len(self.point_color))]
                self.axes.plot(real(self.points[k][1]),
                               -real(self.points[k][0]),
                               real(self.points[k][2]),
                               alpha=self.point_alpha[k],
                               zdir='z', color=color)

    def plot_annotations(self):
        # -X and Y data are switched for plotting purposes
        for annotation in self.annotations:
            vec = annotation['position']
            opts = {'fontsize': self.font_size,
                    'color': self.font_color,
                    'horizontalalignment': 'center',
                    'verticalalignment': 'center'}
            opts.update(annotation['opts'])
            self.axes.text(vec[1], -vec[0], vec[2],
                           annotation['text'], **opts)

    def plot_lines(self):
        for line, fmt, kw in self._lines:
            self.axes.plot(line[0], line[1], line[2], fmt, **kw)

    def plot_arcs(self):
        for arc, fmt, kw in self._arcs:
            self.axes.plot(arc[1, :], -arc[0, :], arc[2, :], fmt, **kw)

    def show(self):
        """
        Display Bloch sphere and corresponding data sets.

        Notes
        -----

        When using inline plotting in Jupyter notebooks, any figure created
        in a notebook cell is displayed after the cell executes. Thus if you
        create a figure yourself and use it create a Bloch sphere with
        ``b = Bloch(..., fig=fig)`` and then call ``b.show()`` in the same
        cell, then the figure will be displayed twice. If you do create your
        own figure, the simplest solution to this is to not call ``.show()``
        in the cell you create the figure in.
        """
        self.render()
        if self.run_from_ipython():
            display(self.fig)
        else:
            self.fig.show()

    def save(self, name=None, format='png', dirc=None, dpin=None):
        """Saves Bloch sphere to file of type ``format`` in directory ``dirc``.

        Parameters
        ----------

        name : str
            Name of saved image. Must include path and format as well.
            i.e. '/Users/Paul/Desktop/bloch.png'
            This overrides the 'format' and 'dirc' arguments.
        format : str
            Format of output image.
        dirc : str
            Directory for output images. Defaults to current working directory.
        dpin : int
            Resolution in dots per inch.

        Returns
        -------
        File containing plot of Bloch sphere.

        """
        self.render()
        # Conditional variable for first argument to savefig
        # that is set in subsequent if-elses
        complete_path = ""
        if dirc:
            if not os.path.isdir(os.getcwd() + "/" + str(dirc)):
                os.makedirs(os.getcwd() + "/" + str(dirc))
        if name is None:
            if dirc:
                complete_path = os.getcwd() + "/" + str(dirc) + '/bloch_' \
                                + str(self.savenum) + '.' + format
            else:
                complete_path = os.getcwd() + '/bloch_' + \
                                str(self.savenum) + '.' + format
        else:
            complete_path = name

        if dpin:
            self.fig.savefig(complete_path, dpi=dpin)
        else:
            self.fig.savefig(complete_path)
        self.savenum += 1
        if self.fig:
            plt.close(self.fig)


def _hide_tick_lines_and_labels(axis):
    '''
    Set visible property of ticklines and ticklabels of an axis to False
    '''
    for a in axis.get_ticklines() + axis.get_ticklabels():
        a.set_visible(False)
